/*
 * =============================================================================
 *
 * Application: JSbigCamCli
 * Copyright  : Jan Soldan, Astronomical Institute, v.v.i, 
 * 				Ondrejov, Czech Republic
 * Email      : soldan@sunstel.asu.cas.cz
 *
 * JSbigCamCli software is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as published by 
 * the Free Software Foundation; either version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * JSbigCamCli software is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License 
 * along with this JRT software; if not, write to the Free Software Foundation, 
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 * The JSbigCamCli code may be used for non-commercial purpose only.
 * 
 * 
 * Important note about the FITS package used:
 * -------------------------------------------
 * Name       : net-ivoa-fits-0.1.jar 
 * Version    : 0.96a October 10, 2000
 * Copyright  : Thomas McGlynn 1997-1999.
 * 
 * This code may be used for any purpose, non-commercial or commercial so long 
 * as this copyright notice is retained in the source code or included in or 
 * referred to in any derived software.
 *
 *=============================================================================
 */

package js.dev.cam.sbig.cli;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.Socket;

import js.dev.CloseParams;
import js.dev.CloseResults;
import js.dev.ERROR;
import js.dev.OpenParams;
import js.dev.OpenResults;
import js.dev.cam.BITPIX;
import js.dev.cam.ExposeCcdProgressResults;
import js.dev.cam.GrabImageParams;
import js.dev.cam.GrabImageResults;
import js.dev.cam.IMAGE_TYPE;
import js.dev.cam.ReadoutCcdProgressResults;
import js.dev.cam.sbig.CCD_INFO_REQUEST;
import js.dev.cam.sbig.CCD_REQUEST;
import js.dev.cam.sbig.CFW_COM_PORT;
import js.dev.cam.sbig.CFW_MODEL_SELECT;
import js.dev.cam.sbig.CFW_POSITION;
import js.dev.cam.sbig.CfwCloseDeviceParams;
import js.dev.cam.sbig.CfwCloseDeviceResults;
import js.dev.cam.sbig.CfwGotoParams;
import js.dev.cam.sbig.CfwGotoResults;
import js.dev.cam.sbig.CfwOpenDeviceParams;
import js.dev.cam.sbig.CfwOpenDeviceResults;
import js.dev.cam.sbig.EstablishLinkParams;
import js.dev.cam.sbig.EstablishLinkResults;
import js.dev.cam.sbig.GetCcdInfoParams;
import js.dev.cam.sbig.GetCcdInfoResults0;
import js.dev.cam.sbig.GetDriverInfoParams;
import js.dev.cam.sbig.GetDriverInfoResults0;
import js.dev.cam.sbig.QueryTemperatureStatusParams;
import js.dev.cam.sbig.QueryTemperatureStatusResults;
import js.dev.cam.sbig.READOUT_MODE;
import js.dev.cam.sbig.ReadoutInfo;
import js.dev.cam.sbig.SetTemperatureRegulationParams;
import js.dev.cam.sbig.SetTemperatureRegulationResults;
import js.dev.cam.sbig.TEMPERATURE_REGULATION;
import net.ivoa.fits.Fits;
import net.ivoa.fits.Header;
import net.ivoa.fits.data.ImageData;
import net.ivoa.fits.hdu.ImageHDU;
import net.ivoa.util.BufferedDataOutputStream;

/**
 * This is a very simple command line application which shows how to control
 * the SBIG CCD camera running as an Java server, i.e. through the JSbigCamSrv
 * server application.
 * 
 * Use:
 * java js.dev.cam.sbig.cli.JSbigCamCli
 * localhost         // IP address. Examples: localhost, 192.168.1.10, etc
 * 5000 	         // Server port number > 1024. Example: 5000, etc.
 * PROGRESS_NO       // Show progress bar? Examples: PROGRESS_YES, PROGRESS_NO
 * FITS_YES          // Save image in FITS format? Examples: FITS_YES, FITS_NO
 * CFW_FLAG          // CFW_YES or CFW_NO
 * CFW_PORT          // CFW_COM0, CFW_COM1, ... CFW_COM3
 * CFW type          // CFW_UNKNOWN, CFW2, CFW5, CFW8, CFWL, CFW402, CFW_AUTO, 
 *                      CFW6A, CFW10, CFW10_SERIAL, CFW9, CFWL8, CFWL8G                  
 * CFW device name   // Device name of CFW-10 SA (external); An empty string if an internal CFW is used.
 * CFW position      // Filter wheel position: 1-10 for CFW-10, 1-5 for CFW-8, etc.
 * LIGHT             // Image type. Examples: LIGHT or DARK
 * IMAGING           // CCD request. Examples: IMAGING, TRACKING
 * 1x1               // Binning modes. Examples: 1x1, 2x2, 3x3 or 9x9
 * numOfExps         // Number of exposures. Examples: 1, 20, etc.
 * exposureInSec     // Exposure time in seconds. Examples: 0.1, 3, 300, etc.
 * 
 * @author soldan
 */


public class JSbigCamCli {

	// CCD Setpoint:
	private static final double ccdSetpoint = -5.0;
	
	// FITS path:
	/** The Constant FITS_PATH. */
	private static final String FITS_PATH  = "/data/observations/test/";
	
	// The FITS file names are generated automatically. They are prepended  
	// by the "L_" or "D_" strings followed by exposure start time:
	// Light frame example: L_2007-11-01T14:53:46.681
	// Dark  frame example: D_2007-11-01T14:55:34.231
	
	// Define FITS keywords:
	/** The Constant OBJECT. */
	private static final String OBJECT     = "unknown";
	
	/** The Constant TELESCOP. */
	private static final String TELESCOP   = "AP-900, Newton 300/4.6 + Televue Paracorr";
	
	/** The Constant OBSERVER. */
	private static final String OBSERVER   = "Jan Soldan";
	
	/** The Constant HISTORY. */
	private static final String HISTORY    = "Raw image";
	
	/** The Constant FOCALLEN. */
	private static final String FOCALLEN   = "1363.0";
	
	/** The Constant APTDIA. */
	private static final String APTDIA     = "300.0";
	
	/** The Constant APTAREA. */
	private static final String APTAREA    = "62800.0";
	
	/** The Constant CBLACK. */
	private static final String CBLACK     = "100";
	
	/** The Constant CWHITE. */
	private static final String CWHITE     = "10000";
	
	/** The Constant PEDESTAL. */
	private static final String PEDESTAL   = "100";
	
	/** The Constant DATAMAX. */
	private static final String DATAMAX    = "60000";
	
	/** The Constant SWCREATE. */
	private static final String SWCREATE   = "JSbigCamCli Version 1.0";
	
	/** The Constant SBSTDVER. */
	private static final String SBSTDVER   = "SBFITSEXT Version 1.0";

	// Print error message:
	/**
	 * Bad params.
	 */
	static private void BadParams() {
		System.out.println("Use: java js.dev.cam.sbig.cli.JSbigCamCli " + 
				           "host port PROGRESS_YES/NO FITS_YES/NO " +
				           "CFW_YES/NO CFW_COM_PORT CFW_MODEL CFW_GOTO " +
				           "LIGHT/DARK IMAGING/TRACKING binning numOfExps exposureInSec");
		System.out.println("Example:");
        System.out.println("java js.dev.cam.sbig.cli.JSbigCamCli " + 
                "localhost 5000 PROGRESS_NO FITS_NO CFW_YES CFW_COM0 CFW10_SERIAL 3 LIGHT IMAGING 1x1 1 1");
	}	
	
	/**
	 * Sleep in milliseconds.
	 * @param ms the ms
	 */
	/*
	private static void sleepMS(int ms) {
		try{
			Thread.sleep(ms);
		}catch(InterruptedException e){
		}		
	}
	*/
	
	// Main method:
	/**
	 * The main method.
	 * 
	 * @param argv the arguments
	 */
	public static void main(String argv[]) {
		try {
			
		    // Check number of parameters:
		    if(argv.length < 13) {
                BadParams();
                return;		        
		    }
		    
			// Arg. 1: Host IP address.
			String host = argv[0];
			
			// Arg. 2: Server port number.
			int port = Integer.parseInt(argv[1]);
		
			// Arg. 3: CCD readout progress.
			boolean ccdReadoutProgress;
			if (argv[2].equals("PROGRESS_YES")) {
				ccdReadoutProgress = true;
			} else if (argv[2].equals("PROGRESS_NO")) {
				ccdReadoutProgress = false;
			} else {
				BadParams();
				return;
			}
			
			// Arg. 4: Save image in FITS format.
			boolean fitsFlag;
			if (argv[3].equals("FITS_YES")) {
				fitsFlag = true;
			} else if (argv[3].equals("FITS_NO")) {
				fitsFlag = false;
			} else {
				BadParams();
				return;
			}
			
			// Arg. 5: CFW_YES or CFW_NO flag.
			boolean cfwFlag;
			if (argv[4].equals("CFW_YES")) {
				cfwFlag = true;
			} else if (argv[4].equals("CFW_NO")) {
				cfwFlag = false;
			} else {
				BadParams();
				return;
			}
			
			// Arg. 6: CFW RS-232 port.
			CFW_COM_PORT cfwComPort;
			if (argv[5].equals("CFW_COM0")) {
				cfwComPort = CFW_COM_PORT.CFWPORT_COM0;
			} else if (argv[5].equals("CFW_COM1")) {
				cfwComPort = CFW_COM_PORT.CFWPORT_COM1;
			} else if (argv[5].equals("CFW_COM2")) {
				cfwComPort = CFW_COM_PORT.CFWPORT_COM2;
			} else if (argv[5].equals("CFW_COM3")) {
				cfwComPort = CFW_COM_PORT.CFWPORT_COM3;
			} else {
				BadParams();
				return;
			}
			
			// Arg. 7: CFW model.
			CFW_MODEL_SELECT cfwModel;
			if (argv[6].equals("CFW_UNKNOWN")) {
				cfwModel = CFW_MODEL_SELECT.CFWSEL_UNKNOWN;
			} else if (argv[6].equals("CFW2")) {
				cfwModel = CFW_MODEL_SELECT.CFWSEL_CFW2;
			} else if (argv[6].equals("CFW5")) {
				cfwModel = CFW_MODEL_SELECT.CFWSEL_CFW5;
			} else if (argv[6].equals("CFW8")) {
				cfwModel = CFW_MODEL_SELECT.CFWSEL_CFW8;
			} else if (argv[6].equals("CFWL")) {
				cfwModel = CFW_MODEL_SELECT.CFWSEL_CFWL;
			} else if (argv[6].equals("CFW402")) {
				cfwModel = CFW_MODEL_SELECT.CFWSEL_CFW402;
			} else if (argv[6].equals("CFW_AUTO")) {
				cfwModel = CFW_MODEL_SELECT.CFWSEL_AUTO;
			} else if (argv[6].equals("CFW6A")) {
				cfwModel = CFW_MODEL_SELECT.CFWSEL_CFW6A;
			} else if (argv[6].equals("CFW10")) {
				cfwModel = CFW_MODEL_SELECT.CFWSEL_CFW10;
			} else if (argv[6].equals("CFW10_SERIAL")) {
				cfwModel = CFW_MODEL_SELECT.CFWSEL_CFW10_SERIAL;
			} else if (argv[6].equals("CFW9")) {
				cfwModel = CFW_MODEL_SELECT.CFWSEL_CFW9;
			} else if (argv[6].equals("CFWL8")) {
				cfwModel = CFW_MODEL_SELECT.CFWSEL_CFWL8;
			} else if (argv[6].equals("CFWL8G")) {
				cfwModel = CFW_MODEL_SELECT.CFWSEL_CFWL8G;
			} else {
				BadParams();
				return;
			}
			
			// Arg. 8: CFW GOTO position.
			CFW_POSITION cfwPosition; 
			switch(Integer.parseInt(argv[7])) {
				case 1: 
					cfwPosition = CFW_POSITION.CFWP_1;
					break;
				case 2:
					cfwPosition = CFW_POSITION.CFWP_2;
					break;
				case 3:
					cfwPosition = CFW_POSITION.CFWP_3;
					break;
				case 4:
					cfwPosition = CFW_POSITION.CFWP_4;
					break;
				case 5:
					cfwPosition = CFW_POSITION.CFWP_5;
					break;
				case 6:
					cfwPosition = CFW_POSITION.CFWP_6;
					break;
				case 7:
					cfwPosition = CFW_POSITION.CFWP_7;
					break;
				case 8:
					cfwPosition = CFW_POSITION.CFWP_8;
					break;
				case 9:
					cfwPosition = CFW_POSITION.CFWP_9;
					break;
				case 10:
					cfwPosition = CFW_POSITION.CFWP_10;
					break;
				default:
					BadParams();
					return;
			}
			
			// Arg. 9: Image type. 
			IMAGE_TYPE imageType;
			if (argv[8].equals("LIGHT")) {
				imageType = IMAGE_TYPE.LIGHT;
			} else if (argv[8].equals("DARK")) {
				imageType = IMAGE_TYPE.DARK;
			} else {
				BadParams();
				return;
			}

			// Arg. 10: CCD request. 
			CCD_REQUEST ccdRequest;
			if (argv[9].equals("IMAGING")) {
				ccdRequest = CCD_REQUEST.CCD_IMAGING;
			} else if (argv[9].equals("TRACKING")) {
				ccdRequest = CCD_REQUEST.CCD_TRACKING;
			} else {
				BadParams();
				return;
			}
			
			// Arg. 11: Binning mode.
			READOUT_MODE binning;
			if (argv[10].equals("1x1")) {
				binning = READOUT_MODE.BIN_1x1;
			} else if (argv[10].equals("2x2")) {
				binning = READOUT_MODE.BIN_2x2;
			} else if (argv[10].equals("3x3")) {
				binning = READOUT_MODE.BIN_3x3;
			} else if (argv[10].equals("9x9")) {
				binning = READOUT_MODE.BIN_9x9;
			} else {
				BadParams();
				return;
			}
			
			// Arg. 12: Number of exposures.
			int numExposures = Integer.parseInt(argv[11]);

			// Arg. 13: Exposure time in seconds.
			double exposure = Double.parseDouble(argv[12]);
			
			System.out.println("Host: "                + host               + 
					           ", Port: "              + port               + 
					           ", Progress: "          + ccdReadoutProgress +
					           ", FITS flag: "         + fitsFlag           +
					           ", CFW flag: "          + cfwFlag            +
					           ", CFW_COM_PORT: "      + cfwComPort         +
					           ", CFW model: "         + cfwModel           +
					           ", CFW GOTO: "          + cfwPosition        +
					           ", Image type: "        + imageType          + 
					           ", CCD request: "       + ccdRequest         + 
					           ", Binning: "           + binning            + 
					           ", numExposures: "      + numExposures       +
					           ", exposure time: "     + exposure);
			
			// Create new socket:
			Socket socket = new Socket(host, port);
			ObjectOutputStream os = new ObjectOutputStream(socket.getOutputStream());					
			ObjectInputStream  is = new ObjectInputStream(socket.getInputStream());
			System.out.println("server ok");
						
			// open: 
			// This method opens the driver and device in one call. You can use it 
			// instead of the openDriver & openDevice methods below.
			OpenParams op = new OpenParams("USB1");
			os.writeObject(op);
			os.reset();
			OpenResults or = (OpenResults)is.readObject();				
			if (or.getError() == ERROR.CE_NO_ERROR){
				System.out.println("open                     : ok");
			} else {
				System.out.println("open error               : " + or.getError());
			}
			
			/*
			// openDriver:
			OpenDriverParams odp = new OpenDriverParams();
			os.writeObject(odp);
			os.reset();
			OpenDriverResults odr = (OpenDriverResults)is.readObject();				
			if (odr.getError() == PAR_ERROR.CE_NO_ERROR){
				System.out.println("openDriver               : ok");
			} else {
				System.out.println("openDriver error         : " + odr.getError());
			}
			
			// openDevice:
			OpenDeviceParams odep = new OpenDeviceParams();
			odep.setDeviceType(DEVICE_TYPE.DEV_USB1);
			os.writeObject(odep);
			os.reset();
			OpenDeviceResults oder = (OpenDeviceResults)is.readObject();			
			if (oder.getError() == PAR_ERROR.CE_NO_ERROR){
				System.out.println("openDevice               : ok");
			} else {
				System.out.println("openDevice error         : " + oder.getError());				
			}
			*/
			
			// getDriverInfo:
			GetDriverInfoParams gdip = new GetDriverInfoParams();
			os.writeObject(gdip);
			os.reset();
			GetDriverInfoResults0 gdir = (GetDriverInfoResults0)is.readObject();				
			if (gdir.getError() == ERROR.CE_NO_ERROR){
				System.out.println("getDriverInfo            : ok");
				System.out.println("version                  : " + gdir.getVersion());
				System.out.println("name                     : " + gdir.getName());
				System.out.println("maxRequest               : " + gdir.getMaxRequest());
			} else {
				System.out.println("getDriverInfo error      : " + gdir.getError());
			}
					
			// establishLink:
			EstablishLinkParams elp = new EstablishLinkParams();
			os.writeObject(elp);
			os.reset();
			EstablishLinkResults elr = (EstablishLinkResults)is.readObject();					
			if (elr.getError() == ERROR.CE_NO_ERROR) {
				System.out.println("establishLink            : ok");
				System.out.println("Camera type              : " + elr.getCameraType());
			} else {
				System.out.println("establishLink error      : " + elr.getError());
			}

			// setTemperatureRegulation:
			SetTemperatureRegulationParams strp = new SetTemperatureRegulationParams();
			strp.setRegulation(TEMPERATURE_REGULATION.REGULATION_ON);
			strp.setCcdSetpoint(ccdSetpoint);
			os.writeObject(strp);
			os.reset();
			SetTemperatureRegulationResults strr = (SetTemperatureRegulationResults)is.readObject();					
			if (strr.getError() == ERROR.CE_NO_ERROR) {			
				System.out.println("setTemperatureRegulation : ok");
			} else {
				System.out.println("setTemperatureRegulation error : " + strr.getError());				
			}

			// queryTemperatureStatus:
			QueryTemperatureStatusParams qtsp = new QueryTemperatureStatusParams();
			os.writeObject(qtsp);
			os.reset();
			QueryTemperatureStatusResults qtsr = (QueryTemperatureStatusResults)is.readObject();					
			if (qtsr.getError() == ERROR.CE_NO_ERROR) {
				System.out.printf("queryTemperatureStatus   : ok -> ccdSetpoint: %.1f, ccdTemp: %.1f, Power: %.1f\n", qtsr.getCcdSetpoint(), qtsr.getCcdThermistor(), qtsr.getPower() * 100.0);
			}
			
			// activateRelay:
			/*
			ActivateRelayParams arp = new ActivateRelayParams();
			arp.set(100, 0, 100, 0);
			os.writeObject(arp);
			os.reset();
			ActivateRelayResults arr = (ActivateRelayResults)is.readObject();					
			if (arr.getError() == ERROR.CE_NO_ERROR) {
				System.out.println("activateRelay            : ok");
			} else {
				System.out.println("activateRelay            : " + arr.getError());
			}
			
			// queryActivateRelayStatus:
			QueryActivateRelayStatusParams  qarsp = new QueryActivateRelayStatusParams();
			QueryActivateRelayStatusResults qarsr = null;
			do {
				os.writeObject(qarsp);
				os.reset();
				qarsr = (QueryActivateRelayStatusResults)is.readObject();
				System.out.println("activateRelay status     : +X: " 
						           + qarsr.isTxPlus()  + 
						           "  -X: " + qarsr.isTxMinus() +
						           "  +Y: " + qarsr.isTyPlus()  + 
						           "  -Y: " + qarsr.isTyMinus());				
			} while (qarsr.isActivated());			
			*/
			
			// getCcdInfo - CCD_INFO_IMAGING:
			GetCcdInfoParams gccdip = new GetCcdInfoParams();
			gccdip.setRequest(CCD_INFO_REQUEST.CCD_INFO_IMAGING);
			os.writeObject(gccdip);
			os.reset();
			GetCcdInfoResults0 imgCcdir0 = (GetCcdInfoResults0)is.readObject();					
			if (imgCcdir0.getError() == ERROR.CE_NO_ERROR) {			
				System.out.println("getCcdInfo 0             : ok");
				System.out.println("firmwareVersion          : " + imgCcdir0.getFirmwareVersion());
				System.out.println("cameraType               : " + imgCcdir0.getCameraType());
				System.out.println("name                     : " + imgCcdir0.getName());
				System.out.println("readoutModes             : " + imgCcdir0.getReadoutModes());
				int rmMax = imgCcdir0.getReadoutModes();
				for (int rm = 0; rm < rmMax; rm++) {
					ReadoutInfo ri = imgCcdir0.getReadoutInfo()[rm];
					System.out.println("--------------------------------");
					System.out.println("ReadoutInfo              : " + gccdip.getRequest());
					System.out.println("mode                     : " + ri.getMode());
					System.out.println("width                    : " + ri.getWidth());
					System.out.println("height                   : " + ri.getHeight());
					System.out.println("gain                     : " + ri.getGain());
					System.out.println("pixelWidth               : " + ri.getPixelWidth());
					System.out.println("pixelHeight              : " + ri.getPixelHeight());
				}
			} else {
				System.out.println("getCcdInfo 0 error       : " + imgCcdir0.getError());				
			}
			System.out.println("--------------------------------");

			// getCcdInfo - CCD_INFO_TRACKING:
			gccdip.setRequest(CCD_INFO_REQUEST.CCD_INFO_TRACKING);
			os.writeObject(gccdip);
			os.reset();
			GetCcdInfoResults0 trackCcdir0 = (GetCcdInfoResults0)is.readObject();					
			if (trackCcdir0.getError() == ERROR.CE_NO_ERROR) {			
				System.out.println("getCcdInfo 1             : ok");
				System.out.println("firmwareVersion          : " + trackCcdir0.getFirmwareVersion());
				System.out.println("cameraType               : " + trackCcdir0.getCameraType());
				System.out.println("name                     : " + trackCcdir0.getName());
				System.out.println("readoutModes             : " + trackCcdir0.getReadoutModes());
				int rmMax = trackCcdir0.getReadoutModes();
				for (int rm = 0; rm < rmMax; rm++) {
					ReadoutInfo ri = trackCcdir0.getReadoutInfo()[rm];
					System.out.println("--------------------------------");
					System.out.println("ReadoutInfo              : " + gccdip.getRequest());
					System.out.println("mode                     : " + ri.getMode());
					System.out.println("width                    : " + ri.getWidth());
					System.out.println("height                   : " + ri.getHeight());
					System.out.println("gain                     : " + ri.getGain());
					System.out.println("pixelWidth               : " + ri.getPixelWidth());
					System.out.println("pixelHeight              : " + ri.getPixelHeight());
				}				
			} else {
				System.out.println("getCcdInfo 1 error       : " + trackCcdir0.getError());				
			}
			System.out.println("--------------------------------");

			/*
			// getCcdInfo - CCD_INFO_EXTENDED:
			gccdip.setRequest(CCD_INFO_REQUEST.CCD_INFO_EXTENDED);
			os.writeObject(gccdip);
			os.reset();
			GetCcdInfoResults2 gccdir2 = new GetCcdInfoResults2();
			gccdir2 = (GetCcdInfoResults2)is.readObject();					
			if (gccdir2.getError() == ERROR.CE_NO_ERROR) {			
				System.out.println("getCcdInfo 2             : ok");
				System.out.println("badColumns               : " + gccdir2.getBadColumns());
				int [] columns = gccdir2.getColumns();
				for (int c=0; c < gccdir2.getBadColumns(); c++) {
					System.out.println("bad column               : " + columns[c]);
				}
				System.out.println("imagingABG               : " + gccdir2.getImagingAbg());
				System.out.println("serialNumber             : " + gccdir2.getSerialNumber());
			} else {
				System.out.println("getCcdInfo 2 error       : " + gccdir2.getError());				
			}
			System.out.println("--------------------------------");			
			*/
			
			/*
			// getCcdInfo - CCD_INFO_EXTENDED2_IMAGING:
			gccdip.setRequest(CCD_INFO_REQUEST.CCD_INFO_EXTENDED2_IMAGING);
			os.writeObject(gccdip);
			os.reset();
			GetCcdInfoResults4 gccdir4 = new GetCcdInfoResults4();
			gccdir4 = (GetCcdInfoResults4)is.readObject();					
			if (gccdir4.getError() == ERROR.CE_NO_ERROR) {			
				System.out.println("getCcdInfo 4             : ok");
				int capabilitiesBits = gccdir4.getCapabilitiesBits();
				System.out.println("capabilityBits           : " + capabilitiesBits);
				if ((capabilitiesBits & 2) == 0) {
					if ((capabilitiesBits & 1) == 0) {
						System.out.println("CCD                      : Full Frame Device");
					} else {
						System.out.println("CCD                      : Frame Transfer Device");					
					}
					System.out.println("CCD                      : No Electronic Shutter");
				} else {
					System.out.println("CCD                      : Interline Imaging CCD with Electronics Shutter");					
				}
				System.out.println("dumpExtra                : " + gccdir4.getDumpExtra());
			} else {
				System.out.println("getCcdInfo 4 error       : " + gccdir4.getError());				
			}
			System.out.println("--------------------------------");			
			*/
			
			/*
			// getCcdInfo - CCD_INFO_EXTENDED2_TRACKING:
			gccdip.setRequest(CCD_INFO_REQUEST.CCD_INFO_EXTENDED2_TRACKING);
			os.writeObject(gccdip);
			os.reset();
			gccdir4 = (GetCcdInfoResults4)is.readObject();					
			if (gccdir4.getError() == ERROR.CE_NO_ERROR) {			
				System.out.println("getCcdInfo 5             : ok");
				System.out.println("capabilityBits           : " + gccdir4.getCapabilitiesBits());
				int capabilitiesBits = gccdir4.getCapabilitiesBits();
				if ((capabilitiesBits & 2) == 0) {
					if ((capabilitiesBits & 1) == 0) {
						System.out.println("CCD                      : Full Frame Device");
					} else {
						System.out.println("CCD                      : Frame Transfer Device");					
					}
					System.out.println("CCD                      : No Electronic Shutter");
				} else {
					System.out.println("CCD                      : Interline Imaging CCD with Electronics Shutter");					
				}
				System.out.println("dumpExtra                : " + gccdir4.getDumpExtra());
			} else {
				System.out.println("getCcdInfo 5 error       : " + gccdir4.getError());				
			}
			System.out.println("--------------------------------");			
			*/
						
			/*
			// SBIG CFW API:
			CfwParams cfwp = new CfwParams();
			cfwp.setCfwModel(CFW_MODEL_SELECT.CFWSEL_CFW8);
			cfwp.setCfwCommand(CFW_COMMAND.CFWC_INIT);
			
			os.writeObject(cfwp);
			os.reset();
			CfwResults cfwr = (CfwResults)is.readObject();					
			if (cfwr.getError() == ERROR.CE_NO_ERROR) {			
				System.out.println("cfw                      : ok");
			} else {
				System.out.println("cfw                      : " + cfwr.getError());				
			}
			*/

			if (cfwFlag) {

				// CFW - open a serial port device.
				if (cfwModel == CFW_MODEL_SELECT.CFWSEL_CFW10_SERIAL) {
					CfwOpenDeviceParams cfwodp = new CfwOpenDeviceParams();
					cfwodp.setCfwModel(cfwModel);
					cfwodp.setCfwComPort(cfwComPort);
					
					os.writeObject(cfwodp);
					os.reset();
					CfwOpenDeviceResults cfwodr = (CfwOpenDeviceResults)is.readObject();					
					if (cfwodr.getError() == ERROR.CE_NO_ERROR) {			
						System.out.println("cfwOpenDevice            : ok");
					} else {
						System.out.println("cfwOpenDevice            : " + cfwodr.getError());
						return;
					}
				}
			
				// CFW - goto to the requested position:
				CfwGotoParams cfwgtp = new CfwGotoParams();
				cfwgtp.setCfwModel(cfwModel);
				cfwgtp.setCfwPosition(cfwPosition);
			
				os.writeObject(cfwgtp);
				os.reset();
				CfwGotoResults cfwgtr = (CfwGotoResults)is.readObject();					
				if (cfwgtr.getError() == ERROR.CE_NO_ERROR) {			
					System.out.println("cfwGoto                  : ok");
				} else {
					System.out.println("cfwGoto                  : " + cfwgtr.getError());				
				}
			}
			
			// Grab series of images:
			GrabImageParams  gip = new GrabImageParams(
											exposure, 
											imageType, 
											ccdRequest,
											binning,
											ccdReadoutProgress);
			GrabImageResults gir;

			for(int i = 1; i <= numExposures; i++) {
			
				// Send QueryTemperatureStatusParams object to the server:
				os.writeObject(qtsp);
				os.reset();
				// Read QueryTemperatureStatusResults object from the server:
				qtsr = (QueryTemperatureStatusResults)is.readObject();
				if (qtsr.getError() == ERROR.CE_NO_ERROR) {
					System.out.printf("queryTemperatureStatus   : ok -> ccdSetpoint: %.1f, ccdTemp: %.1f, Power: %.1f\n", qtsr.getCcdSetpoint(), qtsr.getCcdThermistor(), qtsr.getPower() * 100.0);
				}
				
				// Send GrabImageParams object to the server:
				os.writeObject(gip);
				os.reset();	
				
				// Monitor GrabImage progress:
				while (true) {
					Object obj = is.readObject();
					
					if (obj instanceof ExposeCcdProgressResults) {
						System.out.println("exposeCcd progress       : " + 
								((ExposeCcdProgressResults)obj).getProgress());					
					} else if (obj instanceof ReadoutCcdProgressResults) {
						System.out.println("readoutCcd progress      : " + 
								((ReadoutCcdProgressResults)obj).getProgress());					
					} else {
						gir = (GrabImageResults)obj;
						break;
					}
				}
				
				if (gir.getError() != ERROR.CE_NO_ERROR) {
					System.out.println("grabImage error          : " + gir.getError());
					break;
				}

				if (fitsFlag) {
					// Save image in FITS format:
					
					// FITS path name:
					String fitsPathName;
			
					if (gip.getImageType() == IMAGE_TYPE.LIGHT) {
						fitsPathName = FITS_PATH + "L_" + gir.getExposureStartTimeString();
					} else {
						fitsPathName = FITS_PATH + "D_" + gir.getExposureStartTimeString();
					}
				
					// Create FITS header:
			        Header header  = new Header();
			        
			        // FITS Standard Mandatory Keywords:
			        header.setSimple(true);
			        header.setBitpix(BITPIX.SHORT.getBitpix());
			        header.setNaxes(2);
			        header.setNaxis(1, gir.getWidth());
			        header.setNaxis(2, gir.getHeight());

			        // FITS Standard Reserved Keywords:
			        header.addValue("OBJECT"  , OBJECT, "");
			        header.addValue("TELESCOP", TELESCOP, "");
			        header.addValue("INSTRUME", imgCcdir0.getName(), "");
			        header.addValue("OBSERVER", OBSERVER, "");
			        header.addValue("DATE-OBS", gir.getExposureStartTimeString(), "");
			        header.addValue("BSCALE"  , 1.0, "");    	
			        header.addValue("BZERO"   , 32768.0, "");
			        header.addValue("HISTORY" , HISTORY, "");
			        
			        // SBIG FITS Mandatory Keywords:
			        header.addValue("EXPTIME" , exposure, "");
			        String ccdTemp = String.format("%.2f", qtsr.getCcdThermistor());
			        
			        header.addValue("CCD-TEMP", Double.parseDouble(ccdTemp), "");

			        ReadoutInfo ri;
			        if (gip.getCcdRequest() == CCD_REQUEST.CCD_IMAGING) {
			        	ri = imgCcdir0.getReadoutInfo()[gip.getReadoutMode().getReadoutMode()];
			        } else {
			        	ri = trackCcdir0.getReadoutInfo()[gip.getReadoutMode().getReadoutMode()];
			        }
			        
			        header.addValue("XPIXSZ"  , ri.getPixelWidth(),  "");
			        header.addValue("YPIXSZ"  , ri.getPixelHeight(), "");
			        
			        header.addValue("XBINNING", ri.getMode() + 1, "");
			        header.addValue("YBINNING", ri.getMode() + 1, "");
			        
			        header.addValue("XORGSUBF", 0, "");
			        header.addValue("YORGSUBF", 0, "");
			        
			        header.addValue("EGAIN"   , ri.getGain(), "");
			        
			        header.addValue("FOCALLEN", Float.parseFloat(FOCALLEN), "");
			        header.addValue("APTDIA"  , Float.parseFloat(APTDIA),   "");
			        header.addValue("APTAREA" , Float.parseFloat(APTAREA),  "");
			        
			        header.addValue("CBLACK"  , Integer.parseInt(CBLACK),   "");
			        header.addValue("CWHITE"  , Integer.parseInt(CWHITE),   "");
			       
			        header.addValue("PEDESTAL", Integer.parseInt(PEDESTAL), "");
			        header.addValue("DATAMAX" , Integer.parseInt(DATAMAX),  "");
			        
			        header.addValue("SWCREATE", SWCREATE, "");
			        header.addValue("SBSTDVER", SBSTDVER, "");
					
			        if (gip.getImageType() == IMAGE_TYPE.LIGHT) {
				        header.addValue("IMAGETYP", "Light Frame", "");
			        } else {
				        header.addValue("IMAGETYP", "Dark Frame", "");			        	
			        }

			        header.insertComment("JSbigCamCli Test Image.");

			        // Create Image Data object:
			        ImageData data = new ImageData(gir.getArray());
			        
			        // Create Image Header Unit:
			        ImageHDU  hdu = new ImageHDU(header, data);
			        
			        // Create FITS object:
			        Fits fits = new Fits();
			        
			        // Fill FITS object:
			        fits.addHDU(hdu);
			        FileOutputStream fos = new FileOutputStream(fitsPathName);
			        BufferedDataOutputStream bdos = new BufferedDataOutputStream(fos);
			        fits.write(bdos);
			        bdos.flush();
			        bdos. close();        
				
				}			
				System.out.println("Frame No                 : " + i);
			}
			
			/*
			// exposeCcd:
			// An example how to expose CCD:
			ExposeCcdParams eccdp = new ExposeCcdParams();
			eccdp.setCcdRequest(CCD_REQUEST.CCD_IMAGING);
			eccdp.setExposureTime(0.1);	// in seconds
			eccdp.setSepCcdFlag(SEP_CCD_FLAG.START_NONE);
			eccdp.setSepExposureTimeFlag(SEP_EXPOSURE_TIME_FLAG.EXP_NONE);		
			eccdp.setEepCcdFlag(EEP_CCD_FLAG.END_NONE);
			eccdp.setImageType(IMAGE_TYPE.LIGHT);	
			eccdp.setProgress(true);	
			os.writeObject(eccdp);
			os.reset();
			while (true) {
				Object obj = is.readObject();
				if (obj instanceof ExposeCcdProgressResults) {
					System.out.println("exposeCcd progress       : " + 
							((ExposeCcdProgressResults)obj).getProgress());					
				} else if (obj instanceof ExposeCcdResults) {
					ExposeCcdResults eccdr = (ExposeCcdResults)obj;
					if (eccdr.getError() == ERROR.CE_NO_ERROR) {			
						System.out.println("exposeCcd                : ok");
					} else {
						System.out.println("exposeCcd error          : " + eccdr.getError());				
					}
					break;
				}
			}
			*/
			
			/*
			// readoutCcd:
			// An example how to readout CCD:
			ReadoutCcdParams rccdp = new ReadoutCcdParams();
			rccdp.setCcdRequest(CCD_REQUEST.CCD_IMAGING);
			rccdp.setReadoutMode(READOUT_MODE.BIN_1x1);
			rccdp.setTop(0);
			rccdp.setLeft(0);
			rccdp.setWidth(0);
			rccdp.setHeight(0);
			rccdp.setBitpix(BITPIX.SHORT);
			rccdp.setProgress(true);	
			os.writeObject(rccdp);
			os.reset();
			while (true) {
				Object obj = is.readObject();
				if (obj instanceof ReadoutCcdProgressResults) {
					System.out.println("readoutCcd progress      : " + 
							((ReadoutCcdProgressResults)obj).getProgress());					
				} else if (obj instanceof ReadoutCcdResults) {
					ReadoutCcdResults rccdr = (ReadoutCcdResults)obj;
					if (rccdr.getError() == ERROR.CE_NO_ERROR) {
						System.out.println("readoutCcd               : ok");					
						SaveFits sf = new SaveFits(
								              rccdr.getBitpix(),
								              rccdr.getWidth(),
								              rccdr.getHeight(),
								              rccdr.getArray(),
											  "/data/observations/test/TestImg.fits");
						try {
							sf.save();
							System.out.println("saveFits                 : ok");
						}
						catch (FitsException e) {
							System.out.println("saveFits error           : " + e);
						}
					} else {
						System.out.println("readoutCcd               : " + rccdr.getError());						
					}				
					break;
				}
			}
			*/
			
			if (cfwFlag) {
				// CFW - close serial port device.
				if (cfwModel == CFW_MODEL_SELECT.CFWSEL_CFW10_SERIAL) {
					CfwCloseDeviceParams cfwcdp = new CfwCloseDeviceParams();
					cfwcdp.setCfwModel(cfwModel);
					cfwcdp.setCfwComPort(cfwComPort);
					
					os.writeObject(cfwcdp);
					os.reset();
					CfwCloseDeviceResults cfwcdr = (CfwCloseDeviceResults)is.readObject();					
					if (cfwcdr.getError() == ERROR.CE_NO_ERROR) {			
						System.out.println("cfwCloseDevice           : ok");
					} else {
						System.out.println("cfwCloseDevice           : " + cfwcdr.getError());
						return;
					}
				}
			}
			
			// close:
			// This method close both, the device and driver. You can use it instead of
			// the closeDevice and closeDriver methods below.
			CloseParams cp = new CloseParams();
			os.writeObject(cp);
			os.reset();
			CloseResults cr = (CloseResults)is.readObject();			
			if (cr.getError() == ERROR.CE_NO_ERROR){
				System.out.println("close                    : ok");
			} else {
				System.out.println("close error              : " + cr.getError());				
			}
			
			/*
			// closeDevice:
			CloseDeviceParams cdp = new CloseDeviceParams();
			os.writeObject(cdp);
			os.reset();
			CloseDeviceResults cdr = (CloseDeviceResults)is.readObject();			
			if (cdr.getError() == PAR_ERROR.CE_NO_ERROR){
				System.out.println("closeDevice              : ok");
			} else {
				System.out.println("closeDevice error        : " + cdr.getError());				
			}
			
			// closeDriver:
			CloseDriverParams cdrp = new CloseDriverParams();
			os.writeObject(cdrp);
			os.reset();
			CloseDriverResults cdrr = (CloseDriverResults)is.readObject();			
			if (cdrr.getError() == PAR_ERROR.CE_NO_ERROR){
				System.out.println("closeDriver              : ok");
			} else {
				System.out.println("closeDriver error        : " + cdrr.getError());				
			}
			*/
			
			// Disconnect from the server:
			socket.close();
			os.close();
			is.close();
		}
		catch (IOException e) {
			System.out.println("SbigCamCli - IOException : " + e);
		}
		catch (ClassNotFoundException e) {
			System.out.println("SbigCamCli - ClassNotFoundException : " + e);
		}	
		catch (Exception e) {
			System.out.println("SbigCamCli - Exception : " + e);
		}	
	}
}
