/*
 * =============================================================================
 *
 * Application: JWebCamCli
 * Copyright  : Jan Soldan, Astronomical Institute, v.v.i, 
 * 				Ondrejov, Czech Republic
 * Email      : soldan@sunstel.asu.cas.cz
 *
 * JWebCamCli software is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as published by 
 * the Free Software Foundation; either version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * JWebCamCli software is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License 
 * along with this JRT software; if not, write to the Free Software Foundation, 
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 * The JWebCamCli code may be used for non-commercial purpose only.
 * 
 * Important note about the FITS package used:
 * -------------------------------------------
 * Name       : net-ivoa-fits-0.1.jar 
 * Version    : 0.96a October 10, 2000
 * Copyright  : Thomas McGlynn 1997-1999.
 * 
 * This code may be used for any purpose, non-commercial or commercial so long 
 * as this copyright notice is retained in the source code or included in or 
 * referred to in any derived software.
 *
 *=============================================================================
 */

package js.dev.cam.web.cli;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.Socket;

import js.dev.CloseParams;
import js.dev.CloseResults;
import js.dev.ERROR;
import js.dev.OpenParams;
import js.dev.OpenResults;
import js.dev.cam.BITPIX;
import js.dev.cam.web.VideoDisableParams;
import js.dev.cam.web.VideoDisableResults;
import js.dev.cam.web.VideoEnableParams;
import js.dev.cam.web.VideoEnableResults;
import js.dev.cam.web.VideoGrabParams;
import js.dev.cam.web.VideoGrabResults;

import net.ivoa.fits.Fits;
import net.ivoa.fits.FitsException;
import net.ivoa.fits.Header;
import net.ivoa.fits.data.ImageData;
import net.ivoa.fits.hdu.ImageHDU;
import net.ivoa.util.BufferedDataOutputStream;

/**
 * This is a very simple command line application which shows how to control
 * the Logitech QuickCam Pro for Notebook web camera running as an Java server, 
 * i.e. through the JWebCamSrv server application.
 *
 * Use:
 * java js.dev.cam.web.cli.JWebCamCli
 * 		localhost       // IP address. Examples: localhost, 192.168.1.10, etc
 *      5000 	        // Server port number > 1024. Example: 5000, etc.
 *      FITS_YES        // Save image in FITS format? Examples: FITS_YES, FITS_NO
 *      width           // Image width in pixels
 *      height          // Image height in pixels
 *      numOfExposures  // Number of exposures. Examples: 1,2,3,... etc.
 *      
 * @author soldan
 */

public class JWebCamCli {

	// FITS path:
	private static final String FITS_PATH  = "/data/observations/test/";
	
	// The FITS file names are generated automatically. They are prepended  
	// by the "L_" string followed by exposure start time:
	// Light frame example: L_2007-11-01T14:53:46.681
	
	// Define FITS keywords:
	private static final String OBJECT       = "unknown";
	private static final String TELESCOP     = "Carl Zeiss Tessar 2.0/3.7";
	private static final String INSTRUME     = "Logitech WEB Camera for Notebooks";
	private static final String OBSERVER     = "Jan Soldan";
	private static final String HISTORY      = "Raw image";
	private static final String CBLACK       = "100";
	private static final String CWHITE       = "10000";
	private static final String PEDESTAL     = "100";
	private static final String DATAMAX      = "60000";
	private static final String SWCREATE     = "JWebCamCli Version 1.0";
	private static final String SBSTDVER     = "SBFITSEXT Version 1.0";
	
	// Unfortunately we have no information about the following values:
	private static final String EXPTIME      = "0.1";
	private static final String CCD_TEMP     = "0.0";
	private static final String EGAIN        = "1";
	private static final String PIXEL_WIDTH  = "6.4";
	private static final String PIXEL_HEIGHT = "6.4";
	private static final String FOCALLEN     = "1363.0";
	private static final String APTDIA       = "300.0";
	private static final String APTAREA      = "62800.0";

	// Print error message:
	static private void BadParams() {
		System.out.println("Use: java js.dev.cam.web.cli.JWebCamCli " + 
				           "hostIpAddress listenerPort deviceName "    +
				           "fitsSupport imgWidth, imgHeight, numberOfExposures");
		System.out.println("Example:");
        System.out.println("java js.dev.cam.web.cli.JWebCamCli " + 
                		   "localhost 5001 /dev/video1 FITS_NO 320 240 1");
	}	
	
	/**
	 * The msSleep method puts to sleep the current thread for requested milliseconds.
	 * @param ms the ms to set.
	 */
	private static void msSleep(int ms) {
		try{
			Thread.sleep(ms);
		}catch(InterruptedException e){
		}		
	}
	
	private static void saveFits(VideoGrabResults vgr) throws 
								 FitsException, IOException{
		// FITS path name:
		String fitsPathName;

		fitsPathName = FITS_PATH + "L_" + vgr.getExposureStartTimeString();
	
		// Create FITS header:
		Header header  = new Header();
        
		// FITS Standard Mandatory Keywords:
		header.setSimple(true);
		header.setBitpix(BITPIX.SHORT.getBitpix());
		header.setNaxes(2);
		header.setNaxis(1, vgr.getWidth());
		header.setNaxis(2, vgr.getHeight());

		// FITS Standard Reserved Keywords:
		header.addValue("OBJECT"  , OBJECT, "");
		header.addValue("TELESCOP", TELESCOP, "");
		header.addValue("INSTRUME", INSTRUME, "");
		header.addValue("OBSERVER", OBSERVER, "");
		header.addValue("DATE-OBS", vgr.getExposureStartTimeString(), "");
		header.addValue("BSCALE"  , 1.0, "");    	
		header.addValue("BZERO"   , 32768.0, "");
		header.addValue("HISTORY" , HISTORY, "");
        
		// SBIG FITS Mandatory Keywords:
		header.addValue("EXPTIME" , EXPTIME,  "");
		header.addValue("CCD-TEMP", CCD_TEMP, "");

		header.addValue("XPIXSZ"  , PIXEL_WIDTH,  "");
		header.addValue("YPIXSZ"  , PIXEL_HEIGHT, "");
		header.addValue("XBINNING", "1", "");
		header.addValue("YBINNING", "1", "");
        
		header.addValue("XORGSUBF", 0, "");
		header.addValue("YORGSUBF", 0, "");
        
		header.addValue("EGAIN"   , EGAIN, "");
        
		header.addValue("FOCALLEN", Float.parseFloat(FOCALLEN), "");
		header.addValue("APTDIA"  , Float.parseFloat(APTDIA),   "");
		header.addValue("APTAREA" , Float.parseFloat(APTAREA),  "");
        
		header.addValue("CBLACK"  , Integer.parseInt(CBLACK),   "");
		header.addValue("CWHITE"  , Integer.parseInt(CWHITE),   "");
       
		header.addValue("PEDESTAL", Integer.parseInt(PEDESTAL), "");
		header.addValue("DATAMAX" , Integer.parseInt(DATAMAX),  "");
        
		header.addValue("SWCREATE", SWCREATE, "");
		header.addValue("SBSTDVER", SBSTDVER, "");
		
		header.addValue("IMAGETYP", "Light Frame", "");
		header.insertComment("JWebCamCli Test Image.");

		// Create Image Data object:
		ImageData data = new ImageData(vgr.getArray());
        
		// Create Image Header Unit:
		ImageHDU  hdu = new ImageHDU(header, data);
        
		// Create FITS object:
		Fits fits = new Fits();
        
		// Fill FITS object:
		fits.addHDU(hdu);
		FileOutputStream fos = new FileOutputStream(fitsPathName);
		BufferedDataOutputStream bdos = new BufferedDataOutputStream(fos);
		fits.write(bdos);
		bdos.flush();
		bdos. close();        
		System.out.println("saveFits                 : " + fitsPathName);
	}
	
	// Main method:
	public static void main(String argv[]) {
		try {
			
		    // Check number of parameters:
		    if(argv.length < 6) {
                BadParams();
                return;		        
		    }
		    
			// Arg. 1: Host IP address.
			String host = argv[0];
			
			// Arg. 2: Port number.
			int port = Integer.parseInt(argv[1]);
		
			// Arg. 3: Device name.
			String device = argv[2];

			// Arg. 4: FITS support.
			boolean bSaveFits;
			if (argv[3].equals("FITS_YES")) {
				bSaveFits = true;
			} else if (argv[3].equals("FITS_NO")) {
				bSaveFits = false;
			} else {
				BadParams();
				return;
			}
			
			// Arg. 5: Image width.
			int width = Integer.parseInt(argv[4]);
			
			// Arg. 6: Image height.
			int height = Integer.parseInt(argv[5]);
						
			// Arg. 7: Number of exposures.
			int numExposures = Integer.parseInt(argv[6]);
			
			System.out.println("Host: "            + host      + 
					           ", Port: "          + port      + 
					           ", Device: "        + device    + 
					           ", FITS: "          + bSaveFits +
					           ", Width: "         + width     + 
					           ", Height: "        + height    + 
					           ", numExposures: "  + numExposures);
			
			// Create new socket:
			Socket socket = new Socket(host, port);
			ObjectOutputStream os = new ObjectOutputStream(socket.getOutputStream());					
			ObjectInputStream  is = new ObjectInputStream(socket.getInputStream());
			System.out.println("server                   : ok");
						
			// open device: 
			OpenParams op = new OpenParams(device);
			os.writeObject(op);
			os.reset();
			OpenResults or = (OpenResults)is.readObject();				
			if (or.getError() == ERROR.CE_NO_ERROR){
				System.out.println("open                     : ok");
			} else {
				System.out.println("open error               : " + or.getError());
			}

			/*
			// videoGetControls: 
			VideoGetControlsParams vgcp = new VideoGetControlsParams();
			os.writeObject(vgcp);
			os.reset();
			VideoGetControlsResults vgcr = (VideoGetControlsResults)is.readObject();				
			if (vgcr.getError() == ERROR.CE_NO_ERROR){
				System.out.println("videoGetControls         : ok");
				
				if(vgcr.isBBrightness()) {
					System.out.println("Brightness Min           : " + vgcr.getMinBrightness());
					System.out.println("Brightness Max           : " + vgcr.getMaxBrightness());
					System.out.println("Brightness Step          : " + vgcr.getStepBrightness());
					System.out.println("Brightness Def           : " + vgcr.getDefBrightness());
				}
				
				if(vgcr.isBContrast()) {
					System.out.println("Contrast Min             : " + vgcr.getMinContrast());
					System.out.println("Contrast Max             : " + vgcr.getMaxContrast());
					System.out.println("Contrast Step            : " + vgcr.getStepContrast());
					System.out.println("Contrast Def             : " + vgcr.getDefContrast());
				}
				
				if(vgcr.isBSaturation()) {
					System.out.println("Saturation Min           : " + vgcr.getMinSaturation());
					System.out.println("Saturation Max           : " + vgcr.getMaxSaturation());
					System.out.println("Saturation Step          : " + vgcr.getStepSaturation());
					System.out.println("Saturation Def           : " + vgcr.getDefSaturation());	
				}

				if(vgcr.isBSaturation()) {
					System.out.println("Gain Min                 : " + vgcr.getMinGain());
					System.out.println("Gain Max                 : " + vgcr.getMaxGain());
					System.out.println("Gain Step                : " + vgcr.getStepGain());
					System.out.println("Gain Def                 : " + vgcr.getDefGain());	
				}
			} else {
				System.out.println("videoGetControls         : " + vgcr.getError());
			}
			*/
			
			/*
			// videoSetBrightness: 
			VideoSetBrightnessParams vsbp = new VideoSetBrightnessParams(128);
			os.writeObject(vsbp);
			os.reset();
			VideoSetBrightnessResults vsbr = (VideoSetBrightnessResults)is.readObject();				
			if (vsbr.getError() == ERROR.CE_NO_ERROR){
				System.out.println("videoSetBrightness       : ok");
			} else {
				System.out.println("videoSetBrightness       : " + vsbr.getError());
			}
			*/
			
			/*
			// videoSetContrast: 
			VideoSetContrastParams vscp = new VideoSetContrastParams(32);
			os.writeObject(vscp);
			os.reset();
			VideoSetContrastResults vscr = (VideoSetContrastResults)is.readObject();				
			if (vscr.getError() == ERROR.CE_NO_ERROR){
				System.out.println("videoSetContrast         : ok");
			} else {
				System.out.println("videoSetContrast         : " + vscr.getError());
			}
			*/
			
			/*
			// videoSetSaturation: 
			VideoSetSaturationParams vssp = new VideoSetSaturationParams(16);
			os.writeObject(vssp);
			os.reset();
			VideoSetSaturationResults vssr = (VideoSetSaturationResults)is.readObject();				
			if (vscr.getError() == ERROR.CE_NO_ERROR){
				System.out.println("videoSetSaturation       : ok");
			} else {
				System.out.println("videoSetSaturation       : " + vssr.getError());
			}
			*/
			
			/*
			// videoSetGain: 
			VideoSetGainParams vsgp = new VideoSetGainParams(64);
			os.writeObject(vsgp);
			os.reset();
			VideoSetGainResults vsgr = (VideoSetGainResults)is.readObject();				
			if (vsgr.getError() == ERROR.CE_NO_ERROR){
				System.out.println("videoSetGain             : ok");
			} else {
				System.out.println("videoSetGain             : " + vsgr.getError());
			}
			*/
			
			/*
			// videoTryFormat:
			VideoTryFormatResults tfr = new VideoTryFormatResults();
			VideoTryFormatParams  tfp = new VideoTryFormatParams(310, 230);
			os.writeObject(tfp);
			os.reset();
			tfr = (VideoTryFormatResults)is.readObject();
			System.out.println    ("Format                   : " + tfr.getWidth() + " x " + tfr.getHeight());

			tfp = new VideoTryFormatParams(630, 470);
			os.writeObject(tfp);
			os.reset();
			tfr = (VideoTryFormatResults)is.readObject();
			System.out.println    ("Format                   : " + tfr.getWidth() + " x " + tfr.getHeight());
			
			tfp = new VideoTryFormatParams(790, 590);
			os.writeObject(tfp);
			os.reset();
			tfr = (VideoTryFormatResults)is.readObject();
			System.out.println    ("Format                   : " + tfr.getWidth() + " x " + tfr.getHeight());
			
			tfp = new VideoTryFormatParams(1590, 1190);
			os.writeObject(tfp);
			os.reset();
			tfr = (VideoTryFormatResults)is.readObject();
			System.out.println    ("Format                   : " + tfr.getWidth() + " x " + tfr.getHeight());
			*/
			
			// videoEnable:
			VideoEnableParams vep = new VideoEnableParams(width, height);
			os.writeObject(vep);
			os.reset();
			VideoEnableResults ver = (VideoEnableResults)is.readObject();				
			if (ver.getError() == ERROR.CE_NO_ERROR){
				System.out.println("videoEnable              : ok");
			} else {
				System.out.println("videoEnable error        : " + ver.getError());				
			}
						
			// videoGrab:
			VideoGrabParams  vgp = new VideoGrabParams(
					                        width, 
					                        height, 
					                        new short [height][width]);
			
			VideoGrabResults vgr = new VideoGrabResults();
			int ms = 0;
			
			// Grab series of video images.
			for(int i = 1; i <= numExposures; i++) {
				os.writeObject(vgp);
				os.reset();	
				vgr = (VideoGrabResults)is.readObject();				
				if (vgr.getError() == ERROR.CE_NO_ERROR) {
					if (bSaveFits) { saveFits(vgr); }
				} else {
					System.out.println("videoGrab error          : " + vgr.getError());
				}			
				
				System.out.println("Frame No:                : " + i);

				// Sleep between subsequent images if required.
				if (ms > 0) { msSleep(ms); }
			}
			
			/*
			// We may change the image size in the subsequent calls of grabVideo method:
			width  = 320;
			height = 240;
			array = new short [height][width];
			vgp = new VideoGrabParams(width, height, array);
			
			// Grab series of video images.
			for(int i = 0 ; i < numExposures; i++) {
				os.writeObject(vgp);
				os.reset();	
				vgr = (VideoGrabResults)is.readObject();				
				if (vgr.getError() == ERROR.CE_NO_ERROR) {
					if (bSaveFits) { saveFits(vgp, array); }
				} else {
					System.out.println("videoGrab error          : " + vgr.getError());
				}
			}
			*/
			
			// Stop video:
			VideoDisableParams stvp = new VideoDisableParams();
			os.writeObject(stvp);
			os.reset();
			VideoDisableResults stvr = (VideoDisableResults)is.readObject();				
			if (stvr.getError() == ERROR.CE_NO_ERROR){
				System.out.println("stopVideo                : ok");
			} else {
				System.out.println("stopVideo error          : " + stvr.getError());				
			}
			
			// close device:
			CloseParams cp = new CloseParams();
			os.writeObject(cp);
			os.reset();
			CloseResults cr = (CloseResults)is.readObject();			
			if (cr.getError() == ERROR.CE_NO_ERROR){
				System.out.println("close                    : ok");
			} else {
				System.out.println("close error              : " + cr.getError());				
			}
			
			// Disconnect from the server:
			socket.close();
			os.close();
			is.close();
		}
		catch (IOException e) {
			System.out.println("WebCamCli - IOException : " + e);
		}
		catch (ClassNotFoundException e) {
			System.out.println("WebCamCli - ClassNotFoundException : " + e);
		}	
		catch (Exception e) {
			System.out.println("WebCamCli - Exception : " + e);
		}	
	}
}
